<?php

declare(strict_types = 1);

namespace Amazingcat\BraintreeVault\Controller\Card;

use Amazingcat\BraintreeVault\Controller\CardAction;
use Braintree\CreditCard;
use Braintree\Customer;
use Braintree_CustomerSearch;
use Exception;
use Magento\Braintree\Gateway\Config\Config as BraintreeConfig;
use Magento\Customer\Model\Session;
use Magento\Framework\App\Action\Context;
use Magento\Framework\App\ResponseInterface;
use Magento\Framework\Encryption\EncryptorInterface;
use Magento\Framework\Serialize\SerializerInterface;
use Magento\Framework\Stdlib\DateTime\DateTime;
use Magento\Framework\Data\Form\FormKey\Validator;
use Magento\Vault\Api\Data\PaymentTokenFactoryInterface;
use Magento\Vault\Api\Data\PaymentTokenInterface;
use Magento\Vault\Api\PaymentTokenRepositoryInterface;
use Magento\Vault\Model\PaymentTokenManagement;

/**
 * Class CreateAction
 * @package Amazingcat\BraintreeVault\Controller\Card
 */
class CreateAction extends CardAction
{
    /**
     * @var EncryptorInterface
     */
    private $encryptor;

    /**
     * @var PaymentTokenFactoryInterface
     */
    private $paymentTokenFactory;

    /**
     * @param Context $context
     * @param Session $customerSession
     * @param SerializerInterface $serializer
     * @param DateTime $dateTime
     * @param Validator $fkValidator
     * @param EncryptorInterface $encryptor
     * @param PaymentTokenFactoryInterface $paymentTokenFactory
     * @param PaymentTokenRepositoryInterface $tokenRepository
     * @param PaymentTokenManagement $paymentTokenManagement
     * @param BraintreeConfig $braintreeConfig
     */
    public function __construct(
        Context $context,
        Session $customerSession,
        SerializerInterface $serializer,
        DateTime $dateTime,
        Validator $fkValidator,
        EncryptorInterface $encryptor,
        PaymentTokenFactoryInterface $paymentTokenFactory,
        PaymentTokenRepositoryInterface $tokenRepository,
        PaymentTokenManagement $paymentTokenManagement,
        BraintreeConfig $braintreeConfig
    ) {
        parent::__construct(
            $context,
            $customerSession,
            $serializer,
            $dateTime,
            $fkValidator,
            $tokenRepository,
            $paymentTokenManagement,
            $braintreeConfig
        );

        $this->encryptor = $encryptor;
        $this->paymentTokenFactory = $paymentTokenFactory;

        $this->errorsMap[self::ACTION_EXCEPTION] = __('Create failure. Please try again.');
    }

    /**
     * Dispatch request
     *
     * @return ResponseInterface
     */
    public function execute(): ResponseInterface
    {
        $requestValidationResult = $this->isValidRequest();
        if ($requestValidationResult !== self::OK) {
            return $this->createErrorResponse($requestValidationResult);
        }

        $payloadNonce = $this->getPayloadNonce();

        try {
            $customerEmail = $this->customer->getEmail();
            $this->braintreeGateway->paymentMethodNonce()->find($payloadNonce);

            $braintreeCustomers = $this->braintreeGateway->customer()->search([
                Braintree_CustomerSearch::email()->is($customerEmail)
            ]);

            $braintreeCustomerId = $braintreeCustomers->maximumCount() === 0
                ? $this->createBraintreeCustomer()->id
                : $braintreeCustomers->firstItem()->id;

            $result = $this->braintreeGateway->paymentMethod()->create(
                [
                    'customerId' => $braintreeCustomerId,
                    'paymentMethodNonce' => $payloadNonce
                ]
            );

            $this->createPaymentToken($result->paymentMethod);
        } catch (Exception $e) {
            return $this->createErrorResponse(self::ACTION_EXCEPTION);
        }

        return $this->createSuccessMessage();
    }

    /**
     * @return Customer
     */
    private function createBraintreeCustomer(): Customer
    {
        $newBraintreeCustomer = $this->braintreeGateway->customer()->create([
            'firstName' => $this->customer->getFirstname(),
            'lastName' => $this->customer->getLastname(),
            'email' => $this->customer->getEmail()
        ]);

        return $newBraintreeCustomer->customer;
    }

    /**
     * @param CreditCard $result
     * @return PaymentTokenInterface
     */
    private function createPaymentToken(CreditCard $result): PaymentTokenInterface
    {
        $details = [
            'type' => self::CARD_TYPES[$result->cardType],
            'maskedCC' => $result->last4,
            'expirationDate' => $result->expirationDate
        ];

        $paymentToken = $this->paymentTokenFactory->create(PaymentTokenFactoryInterface::TOKEN_TYPE_CREDIT_CARD);
        $paymentToken->setPaymentMethodCode(self::METHOD_CODE)
            ->setCustomerId($this->customer->getId())
            ->setCreatedAt($this->dateTime->date(null, $result->createdAt->getTimestamp()))
            ->setExpiresAt($this->dateTime->date(null, "$result->expirationYear-$result->expirationMonth-01"))
            ->setGatewayToken($result->token)
            ->setTokenDetails($this->serializer->serialize($details))
            ->setIsActive(true)
            ->setIsVisible(true);

        $paymentToken->setPublicHash($this->generatePublicHash($paymentToken));

        return $this->tokenRepository->save($paymentToken);
    }

    /**
     * @param PaymentTokenInterface $paymentToken
     * @return String
     */
    private function generatePublicHash(PaymentTokenInterface $paymentToken): String
    {
        $hashKey = $paymentToken->getGatewayToken();
        if ($paymentToken->getCustomerId()) {
            $hashKey = $paymentToken->getCustomerId();
        }

        $hashKey .= $paymentToken->getPaymentMethodCode()
            . $paymentToken->getType()
            . $paymentToken->getTokenDetails();

        return $this->encryptor->getHash($hashKey);
    }
}
