<?php

declare(strict_types = 1);

namespace Amazingcat\BraintreeVault\Controller\Card;

use Amazingcat\BraintreeVault\Controller\CardAction;
use Braintree\CreditCard;
use Exception;
use Magento\Braintree\Gateway\Config\Config as BraintreeConfig;
use Magento\Customer\Model\Session;
use Magento\Framework\App\Action\Context;
use Magento\Framework\App\ResponseInterface;
use Magento\Framework\Data\Form\FormKey\Validator;
use Magento\Framework\Serialize\SerializerInterface;
use Magento\Framework\Stdlib\DateTime\DateTime;
use Magento\Vault\Api\Data\PaymentTokenInterface;
use Magento\Vault\Api\PaymentTokenRepositoryInterface;
use Magento\Vault\Model\PaymentTokenManagement;

/**
 * Class UpdateAction
 * @package Amazingcat\BraintreeVault\Controller\Card
 */
class UpdateAction extends CardAction
{
    /**
     * @param Context $context
     * @param Session $customerSession
     * @param SerializerInterface $serializer
     * @param DateTime $dateTime
     * @param Validator $fkValidator
     * @param PaymentTokenRepositoryInterface $tokenRepository
     * @param PaymentTokenManagement $paymentTokenManagement
     * @param BraintreeConfig $braintreeConfig
     */
    public function __construct(
        Context $context,
        Session $customerSession,
        SerializerInterface $serializer,
        DateTime $dateTime,
        Validator $fkValidator,
        PaymentTokenRepositoryInterface $tokenRepository,
        PaymentTokenManagement $paymentTokenManagement,
        BraintreeConfig $braintreeConfig
    ) {
        parent::__construct(
            $context,
            $customerSession,
            $serializer,
            $dateTime,
            $fkValidator,
            $tokenRepository,
            $paymentTokenManagement,
            $braintreeConfig
        );

        $this->errorsMap[self::ACTION_EXCEPTION] = __('Update failure. Please try again.');
    }

    /**
     * Dispatch request
     *
     * @return ResponseInterface
     */
    public function execute(): ResponseInterface
    {
        $requestValidationResult = $this->isValidRequest();
        if ($requestValidationResult !== self::OK) {
            return $this->createErrorResponse($requestValidationResult);
        }

        $payloadNonce = $this->getPayloadNonce();
        $publicHash = $this->getRequestInstance()->getPostValue('cc_public_hash');

        if (empty($publicHash)) {
            return $this->createErrorResponse(self::WRONG_TOKEN);
        }

        $paymentToken = $this->paymentTokenManagement->getByPublicHash($publicHash, $this->customer->getId());
        if (is_null($paymentToken)) {
            return $this->createErrorResponse(self::WRONG_TOKEN);
        }

        try {
            $this->braintreeGateway->paymentMethodNonce()->find($payloadNonce);

            $result = $this->braintreeGateway->paymentMethod()->update(
                $paymentToken->getGatewayToken(),
                [
                    'paymentMethodNonce' => $payloadNonce
                ]
            );

            $this->updatePaymentToken($result->paymentMethod, $paymentToken);
        } catch (Exception $e) {
            return $this->createErrorResponse(self::ACTION_EXCEPTION);
        }

        return $this->createSuccessMessage();
    }

    /**
     * @param CreditCard $result
     * @param PaymentTokenInterface $paymentToken
     * @return PaymentTokenInterface
     */
    private function updatePaymentToken(CreditCard $result, PaymentTokenInterface $paymentToken): PaymentTokenInterface
    {
        $details = [
            'type' => self::CARD_TYPES[$result->cardType],
            'maskedCC' => $result->last4,
            'expirationDate' => $result->expirationDate
        ];

        $paymentToken->setExpiresAt($this->dateTime->date(null, "$result->expirationYear-$result->expirationMonth-01"))
            ->setTokenDetails($this->serializer->serialize($details));

        return $this->tokenRepository->save($paymentToken);
    }
}
